/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "common.h"

/* This is a thread implementation for Win32 with quick hack        */

/* We need this grobal for cheking if initialization is finished.   */
int blas_server_avail = 0;

/* Local Variables */
static unsigned long server_lock       = 0;

static blas_pool_t   pool;
static HANDLE	    blas_threads   [MAX_CPU_NUMBER];
static DWORD	    blas_threads_id[MAX_CPU_NUMBER];

/* This is a main routine of threads. Each thread waits until job is */
/* queued.                                                           */

static DWORD WINAPI blas_thread_server(void *arg){

  /* Thread identifier */
  long  cpu = (long)arg;
  void *buffer;
  blas_queue_t	*queue;
  DWORD action;
  HANDLE handles[] = {pool.filled, pool.killed};
#ifdef USE_BUSYWAIT
  int busywaitdriven;
#endif

  /* Each server needs each buffer */
  buffer   = blas_memory_alloc(2);

#ifdef SMP_DEBUG
    fprintf(STDERR, "Server[%2ld] Thread is started!\n", cpu);
#endif

  while (1){

 /* Waiting for Queue */

#ifdef SMP_DEBUG
    fprintf(STDERR, "Server[%2ld] Waiting for Queue.\n", cpu);
#endif

    do {
      action = WaitForMultipleObjects(2, handles, FALSE, INFINITE);
    } while ((action != WAIT_OBJECT_0) && (action == WAIT_OBJECT_0 + 1));

    if (action == WAIT_OBJECT_0 + 1) break;

#ifdef SMP_DEBUG
    fprintf(STDERR, "Server[%2ld] Got it.\n", cpu);
#endif

    EnterCriticalSection(&pool.lock);

    queue = pool.queue;
    if (queue) pool.queue = queue->next;

    LeaveCriticalSection(&pool.lock);

    if (queue)  {
      if (pool.queue) SetEvent(pool.filled);

      int mode          = queue -> mode;
      long m            = queue -> m;
      long n            = queue -> n;
      long k            = queue -> k;
      void *a           = queue -> a;
      void *b           = queue -> b;
      void *c           = queue -> c;
      long lda          = queue -> lda;
      long ldb          = queue -> ldb;
      long ldc          = queue -> ldc;
      double alpha_r    = queue -> alpha[0];
      double alpha_i    = queue -> alpha[1];
      void *execroutine = queue -> routine;
      long from         = queue -> from;
      long to           = queue -> to;
      
#ifdef SMP_DEBUG
    fprintf(STDERR, "Server[%2ld] Wakeup.  Mode = 0x%03x M = %3ld N=%3ld K=%3ld\n",
	    cpu, queue->mode, queue->m, queue->n, queue->k);
#endif
    
    /* Now starting calculation.  */

    if (!(mode & BLAS_COMPLEX)){
      if (!(mode & BLAS_DOUBLE)){
	/* REAL / Single */
	int (*routine)(long, long, long, float, 
		       float *, long, float *, long, 
		       float *, long, void *, long, long);
	routine = execroutine;
	(routine)(m, n, k, (float)alpha_r, 
		  a, lda, b, ldb, c, ldc, buffer, from, to);
      } else {
	/* REAL / Double */
	int (*routine)(long, long, long, double, 
		       double *, long, double *, long,
		       double *, long, void *, long, long);
	routine = execroutine;
	(routine)(m, n, k, (double)alpha_r,
		  a, lda, b, ldb, c, ldc, buffer, from ,to);
      }
    } else {
      if (!(mode & BLAS_DOUBLE)){
	/* COMPLEX / Single */
	int (*routine)(long, long, long, float, float,
		       float *, long, float *, long,
		       float *, long, void *, long, long);
	routine = execroutine;
	(routine)(m, n, k, (float)alpha_r, (float)alpha_i,
		  a, lda, b, ldb, c, ldc, buffer, from, to);
      } else {
	/* COMPLEX / Double */
	int (*routine)(long, long, long, double, double,
		       double *, long, double *, long,
		       double *, long, void *, long, long);
	routine = execroutine;
	(routine)(m, n, k, (double)alpha_r, (double)alpha_i, 
		  a, lda, b, ldb, c, ldc, buffer, from, to);
      }
    }
    
#ifdef SMP_DEBUG
    fprintf(STDERR, "Server[%2ld] Finished!\n", cpu);
#endif
    
    EnterCriticalSection(&queue->lock);
    
    queue -> status = BLAS_STATUS_FINISHED;
    
    LeaveCriticalSection(&queue->lock);
    
    SetEvent(queue->finish);
    }
  }
  
  /* Shutdown procedure */
  
#ifdef SMP_DEBUG
  fprintf(STDERR, "Server[%2ld] Shutdown!\n",  cpu);
#endif
  
  blas_memory_free(buffer);
  
  return 0;
}

/* Initializing routine */
int blas_thread_init(void){
  long i;

  if (blas_server_avail || (blas_cpu_number <= 1)) return 0;
  
  LOCK_COMMAND(&server_lock);

#ifdef SMP_DEBUG
  fprintf(STDERR, "Initializing Thread(Num. threads = %d)\n", 
	  blas_cpu_number);
#endif

  if (!blas_server_avail){

    blas_pool_init(&pool);

    for(i = 0; i < blas_cpu_number - 1; i++){
      blas_threads[i] = CreateThread(NULL, 128 << 10, 
				     blas_thread_server, (void *)i,
				     0, &blas_threads_id[i]);
    }

    blas_server_avail = 1;
  }

  UNLOCK_COMMAND(&server_lock);

  return 0;
}

/* 
   User can call one of two routines.

     exec_blas_async ... immediately returns after jobs are queued.

     exec_blas       ... returns after jobs are finished.
*/

int exec_blas_async(blas_queue_t *queue){

  blas_queue_t *current = pool.queue;

  EnterCriticalSection(&pool.lock);

  if (current) {
    while (current) current = current -> next;
    current = queue;
  } else {
    pool.queue = queue;
  }

  LeaveCriticalSection(&pool.lock);

  SetEvent(pool.filled);

  return 0;
}

int exec_blas_async_wait(blas_queue_t *queue){

#ifdef SMP_DEBUG
    fprintf(STDERR, "Synchronization Waiting.\n");
#endif

    while (queue){
#ifdef SMP_DEBUG
    fprintf(STDERR, "Waiting Queue ..\n");
#endif

      WaitForSingleObject(queue->finish, INFINITE);
      
      if (queue -> last) break;
      queue = queue -> next;
    }

#ifdef SMP_DEBUG
    fprintf(STDERR, "Completely Done.\n\n");
#endif

  return 0;
}

/* Execute Threads */
int exec_blas(blas_queue_t *queue, void *buffer){

  int exec_flag;

  if (queue == NULL) return 0;

  exec_flag = 0;

  if (queue -> next) {
    exec_blas_async(queue -> next);
    exec_flag = 1;
  }
  
 {
    if (!(queue -> mode & BLAS_COMPLEX)){ /* REAL */
      if (!(queue -> mode & BLAS_DOUBLE)){
	int (*routine)(long, long, long, float, 
		       float *, long, float *, long,
		       float *, long, void *, long, long);
	routine = queue -> routine;
	(routine)(queue -> m, queue -> n, queue -> k,
		  (float)queue -> alpha[0],
		  queue -> a, queue -> lda,
		  queue -> b, queue -> ldb,
		  queue -> c, queue -> ldc, buffer,
		  queue -> from, queue -> to);
      } else {
	int (*routine)(long, long, long, double, 
		       double *, long, double *, long,
		       double *, long, void *, long, long);
	routine = queue -> routine;
	(routine)(queue -> m, queue -> n, queue -> k,
		  (double)queue -> alpha[0],
		  queue -> a, queue -> lda,
		  queue -> b, queue -> ldb,
		  queue -> c, queue -> ldc, buffer,
		  queue -> from, queue -> to);
      }
    } else {
      if (!(queue -> mode & BLAS_DOUBLE)){
	int (*routine)(long, long, long, float, float,
		       float *, long, float *, long,
		       float *, long, void *, long, long);
	routine = queue -> routine;
	(routine)(queue -> m, queue -> n, queue -> k,
		  (float)queue -> alpha[0], (float)queue -> alpha[1],
		  queue -> a, queue -> lda,
		  queue -> b, queue -> ldb,
		  queue -> c, queue -> ldc, buffer,
		  queue -> from, queue -> to);
      } else {
	int (*routine)(long, long, long, double, double,
		       double *, long, double *, long,
		       double *, long, void *, long, long);
	routine = queue -> routine;
	(routine)(queue -> m, queue -> n, queue -> k,
		  (double)queue -> alpha[0], (double)queue -> alpha[1],
		  queue -> a, queue -> lda,
		  queue -> b, queue -> ldb,
		  queue -> c, queue -> ldc, buffer,
		  queue -> from, queue -> to);
      }
    }
  }

  if (exec_flag) exec_blas_async_wait(queue -> next);

  return 0;
}

/* Shutdown procedure, but user don't have to call this routine. The */
/* kernel automatically kill threads.                                */

int blas_thread_shutdown_(void){

  int i;

  if (!blas_server_avail) return 0;
  
  LOCK_COMMAND(&server_lock);

  if (blas_server_avail){

    SetEvent(pool.killed);
    
    for(i = 0; i < blas_cpu_number - 1; i++){
      WaitForSingleObject(blas_threads[i], INFINITE);
    }
    
    blas_server_avail = 0;
  }
  
  UNLOCK_COMMAND(&server_lock);
  
  return 0;
}

