/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "common.h"

static const int divide_rule[][2] = 
  {{ 0,  0},
   { 1,  1}, { 1,  2}, { 1,  3}, { 4,  1},
   { 1,  5}, { 2,  3}, { 1,  7}, { 2,  4},
   { 3,  3}, { 2,  5}, { 1, 11}, { 2,  6},
   { 1, 13}, { 2,  7}, { 3,  5}, { 4,  4},
   { 1, 17}, { 3,  6}, { 1, 19}, { 4,  5},
   { 3,  7}, { 2, 11}, { 1, 23}, { 4,  6},
   { 5,  5}, { 2, 13}, { 3,  9}, { 4,  7},
   { 1, 29}, { 5,  6}, { 1, 31}, { 4,  8},
   { 3, 11}, { 2, 17}, { 5,  7}, { 6,  6},
   { 1, 37}, { 2, 19}, { 3, 13}, { 5,  8},
   { 1, 41}, { 6,  7}, { 1, 43}, { 4, 11},
   { 5,  9}, { 2, 23}, { 1, 47}, { 6,  8},
   { 7,  7}, { 5, 10}, { 3, 17}, { 4, 13},
   { 1, 53}, { 6,  9}, { 5, 11}, { 7,  8},
   { 3, 19}, { 2, 29}, { 1, 59}, { 6, 10},
   { 1, 61}, { 2, 31}, { 7,  9}, { 8,  8},
};

int CNAME(int mode, long m, long n, long k, 
	  double alpha_r, double alpha_i,
	  void *a, long lda,
	  void *b, long ldb, 
	  void *c, long ldc, int (*function)(), void *buffer){

  blas_queue_t queue[MAX_CPU_NUMBER];
  long width, i;
  int num_cpu, calc_type;

  long j, offset;
  int divideA, divideB;
  long   sizeA[MAX_CPU_NUMBER];
  long offsetA[MAX_CPU_NUMBER];
  long   sizeB[MAX_CPU_NUMBER];
  long offsetB[MAX_CPU_NUMBER];

  calc_type = ((mode & BLAS_DOUBLE) != 0) + ((mode & BLAS_COMPLEX) != 0) + 2;
  
  for (i = 0; i < blas_cpu_number; i++) blas_queue_init(&queue[i]);

  divideA = divide_rule[blas_cpu_number][0];
  divideB = divide_rule[blas_cpu_number][1];

  /* divide rule for A */
  i = m;
  num_cpu = 0;
  offset = 0;
  while (i > 0){
    width  = blas_quickdivide(i + divideA - num_cpu - 1,
			      divideA - num_cpu);
    i -= width;
    if (i < 0) width = width + i;
    sizeA  [num_cpu] = width;
    offsetA[num_cpu] = offset;
    offset += width;
    num_cpu ++;
  }
  divideA = num_cpu;

  /* divide rule for B */
  i = n;
  num_cpu = 0;
  offset = 0;
  while (i > 0){
    width  = blas_quickdivide(i + divideB - num_cpu - 1,
			      divideB - num_cpu);
    i -= width;
    if (i < 0) width = width + i;
    sizeB  [num_cpu] = width;
    offsetB[num_cpu] = offset;
    offset += width;
    num_cpu ++;
  }
  divideB = num_cpu;

  num_cpu = 0;

  for (j = 0; j < divideB; j ++){
    for (i = 0; i < divideA; i ++){

      queue[num_cpu].mode = mode;
      queue[num_cpu].routine = function;
      queue[num_cpu].m = sizeA[i];
      queue[num_cpu].n = sizeB[j];
      queue[num_cpu].k = k;
      
      if (!(mode & BLAS_TRANSA_T)){
	queue[num_cpu].a = (void *)((long)a + (offsetA[i] << calc_type));
      } else {
	queue[num_cpu].a = (void *)((long)a + (offsetA[i] * lda << calc_type));
      }

      if (!(mode & BLAS_TRANSB_T)){
	queue[num_cpu].b = (void *)((long)b + (offsetB[j] * ldb << calc_type));
      } else {
	queue[num_cpu].b = (void *)((long)b + (offsetB[j] << calc_type));
      }

      queue[num_cpu].c = (void *)((long)c + ((offsetA[i] + ldc * offsetB[j]) << calc_type));

      queue[num_cpu].lda = lda;
      queue[num_cpu].ldb = ldb;
      queue[num_cpu].ldc = ldc;
      queue[num_cpu].alpha[0] = alpha_r;
      queue[num_cpu].alpha[1] = alpha_i;
      queue[num_cpu].next  = &queue[num_cpu + 1];
      
      num_cpu ++;
    }
  }

  if (num_cpu) {
    queue[num_cpu - 1].next = NULL;
    queue[num_cpu - 1].last = 1;
    
    exec_blas(queue, buffer);

  }
   
  for (i = 0; i < blas_cpu_number; i++) blas_queue_destroy(&queue[i]);

  return 0;
}
