/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "common.h"

#ifdef XDOUBLE
#define ERROR_NAME "XGEMM "
#elif defined(DOUBLE)
#define ERROR_NAME "ZGEMM "
#else
#define ERROR_NAME "CGEMM "
#endif

static int (*gemm[])(long, long, long,
		     FLOAT, FLOAT, 
		     FLOAT *, long,
		     FLOAT* , long,
		     FLOAT *, long, FLOAT *) = {
		       GEMM_NN, GEMM_TN, GEMM_RN, GEMM_CN,
		       GEMM_NT, GEMM_TT, GEMM_RT, GEMM_CT,
		       GEMM_NR, GEMM_TR, GEMM_RR, GEMM_CR,
		       GEMM_NC, GEMM_TC, GEMM_RC, GEMM_CC,
};

int NAME(char *TRANSA, char *TRANSB,
	    blasint *M, blasint *N, blasint *K,
	    FLOAT *ALPHA,
	    FLOAT *a, blasint *ldA,
	    FLOAT *b, blasint *ldB,
	    FLOAT *BETA,
	    FLOAT *c, blasint *ldC){
    
  int transa, transb, nrowa, nrowb;
  blasint info;
  blasint lda, ldb, ldc, m, n, k;
  FLOAT alpha_r, alpha_i;
  FLOAT beta_r, beta_i;
  char transA, transB;
  FLOAT *buffer;

#ifdef SMP
  int mode;
#endif

  lda = *ldA;
  ldb = *ldB;
  ldc = *ldC;
  m   = *M;
  n   = *N;
  k   = *K;
  alpha_r = ALPHA[0];
  alpha_i = ALPHA[1];
  beta_r  = BETA[0];
  beta_i  = BETA[1];
  transA = *TRANSA;
  transB = *TRANSB;

  TOUPPER(transA);
  TOUPPER(transB);

  transa = -1;
  transb = -1;

  if (transA == 'N') transa = 0;
  if (transA == 'T') transa = 1;
  if (transA == 'R') transa = 2;
  if (transA == 'C') transa = 3;

  if (transB == 'N') transb = 0;
  if (transB == 'T') transb = 1;
  if (transB == 'R') transb = 2;
  if (transB == 'C') transb = 3;

  nrowa = m;
  if (transa & 1) nrowa = k;
  nrowb = k;
  if (transb & 1) nrowb = n;

/* Test the input parameters. */
  info = 0;

  if (ldc < m)     info = 13;
  if (ldb < nrowb) info = 10;
  if (lda < nrowa) info = 8; 
  if (k < 0)       info = 5;
  if (n < 0)       info = 4;
  if (m < 0)       info = 3;
  if (transb < 0) info = 2;
  if (transa < 0) info = 1;

  if (info){
    BLASFUNC(xerbla)(ERROR_NAME, &info, sizeof(ERROR_NAME));
    return 0;
  }
 
  if ((m == 0) || (n == 0))  return 0;
  
#ifdef SMP
  if (blas_cpu_number == 0) blas_get_cpu_number();
#ifdef SMP_SERVER
  if (blas_server_avail == 0) blas_thread_init();
#endif
#endif

#ifndef SMP
  if ((beta_r != ONE) || (beta_i != ZERO))
     GEMM_BETA(m, n, 0, beta_r, beta_i, NULL, 0, NULL, 0, c, ldc);
#else

#ifdef DOUBLE
  mode  =  BLAS_DOUBLE | BLAS_COMPLEX;
#else
  mode  =  BLAS_SINGLE | BLAS_COMPLEX; 
#endif  
  
  if ((beta_r != ONE) || (beta_i != ZERO))
    beta_thread(mode, m, n, beta_r, beta_i, c, ldc, (void *)GEMM_BETA);
#endif
  
  if (k == 0) return 0;
  
  if ((alpha_r == ZERO) && (alpha_i == ZERO)) return 0;
  
  buffer = (FLOAT *)blas_memory_alloc(0);

#if 0
  printf("M = %d  N   = %d  K = %d  lda = %d  ldb = %d ldc = %d\n",
	  m, n, k, lda, ldb, ldc);
  printf("M = %p  N = %p  K = %p\n", &m, &n, &k);
  printf("transa = %d  transb = %d alpha_r = %e alpha_i = %e\n", 
	  transa, transb, alpha_r, alpha_i);
  printf("beta_r = %e beta_i = %e\n", beta_r, beta_i);
  printf("A = %p  B  = %p  C = %p\n", a, b, c);
  printf("A = (%e,%e)  B  = (%e,%e)  C = (%e,%e)\n", 
	 *(a + 0), *(a + 1), *(b + 0), *(b + 1), *(c + 0), *(c + 1));
#endif

#ifndef SMP

  (gemm[(transb << 2) | transa])
    (m, n, k, alpha_r, alpha_i, a, lda, b, ldb, c, ldc, buffer);
  
#else

  mode |= (transa << BLAS_TRANSA_SHIFT);
  mode |= (transb << BLAS_TRANSB_SHIFT);

  GEMM_THREAD(mode, m, n, k, alpha_r, alpha_i,
	      a, lda, b, ldb, c, ldc, 
	      (void *)gemm[(transb << 2) | transa], buffer);
#endif

  blas_memory_free(buffer);
  
  return 0;
}
